from typing import List

from qualang_tools.wirer.connectivity.wiring_spec import WiringSpec
from qualang_tools.wirer.instruments import Instruments
from qualang_tools.wirer.wirer.channel_specs import ChannelTemplate
from qualang_tools.wirer.wirer.context_manager_multi_object_temp_attr_setting import MultiObjectTempAttrUpdater


def assign_channels_to_spec(
    spec: WiringSpec,
    instruments: Instruments,
    channel_templates: List[ChannelTemplate],
    same_con: bool = False,
    same_slot: bool = False,
) -> bool:
    candidate_channels = _assign_channels_to_spec(spec, instruments, channel_templates, same_con, same_slot)

    # if candidate channels satisfy all the required channel types
    if len(candidate_channels) == len(channel_templates):
        for channel in candidate_channels:
            # remove candidate channel from stack of available channels
            instruments.used_channels.add(channel)
            instruments.available_channels[type(channel)].remove(channel)
            for element in spec.elements:
                # assign channel to the specified element
                if spec.line_type not in element.channels:
                    element.channels[spec.line_type] = []
                element.channels[spec.line_type].append(channel)

    return len(candidate_channels) == len(channel_templates)


def _assign_channels_to_spec(
    spec: WiringSpec,
    instruments: Instruments,
    channel_templates: List[ChannelTemplate],
    same_con: bool,
    same_slot: bool,
    allocated_channels=None,
):
    """
    Recursive function to find any valid combination of channel allocations
    given a wiring specification, a stack of available channels in the
    instruments setup, and a list of desired channel types.
    """
    if allocated_channels is None:
        allocated_channels = []

    # extract the lead/initial channel type
    target_channel_template = channel_templates[0]

    # filter available channels according to the specification
    available_channels = list(
        filter(
            target_channel_template.make_channel_filter(),  # filter function
            instruments.available_channels.get(type(target_channel_template), []),  # iterable
        )
    )

    candidate_channels = []
    for channel in available_channels:
        # make sure to not re-allocate a channel
        if channel in allocated_channels:
            continue

        candidate_channels = [channel]

        # base case: all channels allocated properly
        if len(channel_templates) == 1:
            break

        # recursive case: allocate remaining channels
        else:
            templates_with_same_instr = [
                template for template in channel_templates[1:] if template.instrument_id == channel.instrument_id
            ]
            with MultiObjectTempAttrUpdater(templates_with_same_instr, con=channel.con, slot=channel.slot):
                # recursively allocate the remaining channels
                subsequent_channels = _assign_channels_to_spec(
                    spec,
                    instruments,
                    channel_templates[1:],
                    same_con,
                    same_slot,
                    allocated_channels=candidate_channels,
                )

            candidate_channels.extend(subsequent_channels)

            # without a candidate channel for every type, try the next available channel
            if len(candidate_channels) == len(channel_templates):
                break

            # otherwise, a successful allocation has been made
            else:
                continue

    return candidate_channels



import time
from typing import Dict, Optional

from qcodes import (
    Instrument,
    Parameter,
    MultiParameter,
)
from qcodes.utils.validators import Numbers, Arrays, Enum
from qm import SimulationConfig, generate_qua_script
from qm.qua import program
from qm import QuantumMachinesManager
from qualang_tools.results import wait_until_job_is_paused
from qualang_tools.results import fetching_tool
from qualang_tools.plot import interrupt_on_close
import matplotlib.pyplot as plt
import numpy as np
import warnings

warnings.filterwarnings("ignore")


# noinspection PyAbstractClass
class OPX(Instrument):
    def __init__(
        self,
        config: Dict,
        name: str = "OPX",
        host: str = None,
        port: int = None,
        cluster_name: str = None,
        octave=None,
        close_other_machines: bool = True,
        unwrap_phase: bool = True,
    ) -> None:
        """
        QCoDeS driver for the OPX.

        :param config: python dict containing the configuration expected by the OPX.
        :param name: The name of the instrument used internally by QCoDeS. Must be unique.
        :param host: IP address of the router to which the OPX is connected.
        :param cluster_name: Name of the cluster as defined in the OPX admin panel (from version QOP220)
        :param port: Port of the OPX or main OPX if working with a cluster.
        :param octave: Octave configuration if an Octave is to be used in this experiment.
        :param close_other_machines: Flag to control if opening a quantum machine will close the existing ones. Default is True. Set to False if multiple a QOP is to be used by multiple users or to run several experiments in parallel.
        """
        super().__init__(name)

        self.qm = None
        self.qm_id = None
        self.qmm = None
        self.close_other_machines = close_other_machines
        self.config = None
        self.result_handles = None
        self.job = None
        self.counter = 0
        self.demod_factor = 1
        self.results = {"names": [], "types": [], "buffers": [], "units": []}
        self.prog_id = None
        self.simulated_wf = {}
        self.unwrap_phase = unwrap_phase
        # Parameter for simulation duration
        self.add_parameter(
            "sim_time",
            unit="ns",
            label="sim_time",
            initial_value=100000,
            vals=Numbers(
                4,
            ),
            get_cmd=None,
            set_cmd=None,
        )
        self.measurement_variables = None
        self.add_parameter(
            "readout_sampling_rate",
            unit="Gs/s",
            initial_value=1,
            vals=Enum(1, 2),
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "readout_pulse_length",
            unit="ns",
            vals=Numbers(16, 1e7),
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_stop",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_start",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_step",
            unit="",
            initial_value=0.1,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_npoints",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_full_list",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis1_axis",
            unit="",
            label="Axis 1",
            parameter_class=GeneratedSetPointsArbitrary,
            full_list=self.axis1_full_list,
            vals=Arrays(shape=(self.axis1_npoints.get_latest,)),
        )
        self.add_parameter(
            "axis2_stop",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis2_start",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis2_step",
            unit="",
            initial_value=0.1,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis2_npoints",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis2_full_list",
            unit="",
            initial_value=0,
            get_cmd=None,
            set_cmd=None,
        )
        self.add_parameter(
            "axis2_axis",
            unit="",
            label="Axis 2",
            parameter_class=GeneratedSetPointsArbitrary,
            full_list=self.axis2_full_list,
            vals=Arrays(shape=(self.axis1_npoints.get_latest,)),
        )
        # Open QMM
        self.connect_to_qmm(host=host, port=port, cluster_name=cluster_name, octave=octave)
        # Set config
        self.set_config(config=config)
        # Open QM
        self.open_qm(close_other_machines)

    def connect_to_qmm(self, host: str = None, port: int = None, cluster_name: str = None, octave=None):
        """
        Enable the connection with the OPX by creating the QuantumMachineManager.
        Displays the connection message with idn when the connection is established.

        :param host: IP address of the router to which the OPX is connected.
        :param port: Port of the OPX or main OPX if working with a cluster.
        :param cluster_name: Name of the cluster as defined in the OPX admin panel (from version QOP220)
        :param octave: Octave configuration if an Octave is to be used in this experiment.
        """
        self.qmm = QuantumMachinesManager(host=host, port=port, cluster_name=cluster_name, octave=octave)
        self.connect_message()

    def connect_message(self, idn_param: str = "IDN", begin_time: Optional[float] = None) -> None:
        """
        Print a standard message on initial connection to an instrument.

        Args:
            idn_param: Name of parameter that returns ID dict.
                Default ``IDN``.
            begin_time: ``time.time()`` when init started.
                Default is ``self._t0``, set at start of ``Instrument.__init__``.
        """
        idn = {"vendor": "Quantum Machines"}
        idn.update(self.get(idn_param))
        # Get OPX model id and corresponding qop version
        if "OPX" in idn.keys():
            model = idn["OPX"][0]
            qop = idn["OPX"]
        elif "server" in idn.keys():
            model = idn["server"][0]
            qop = idn["server"]
        elif "QOP" in idn.keys():
            model = idn["QOP"][0]
            qop = idn["QOP"]
        else:
            model = 0
            qop = ""
        # Get SDK id
        if "qm-qua" in idn.keys():
            sdk = "qm-qua"
        elif "client" in idn.keys():
            sdk = "client"
        else:
            sdk = ""
        # Get OPX model
        if model == "1":
            idn["model"] = "OPX"
        elif model == "2":
            idn["model"] = "OPX+"
        elif model == "3":
            idn["model"] = "OPX1000"
        else:
            idn["model"] = ""
        t = time.time() - (begin_time or self._t0)
        con_msg = (
            f"Connected to: {idn['vendor']} {idn['model']} in {t:.2f}s. "
            f"QOP Version = {qop}, SDK Version = {idn[sdk]}."
        )

        print(con_msg)
        self.log.info(f"Connected to instrument: {idn}")

    def get_idn(self) -> Dict[str, Optional[str]]:
        """
        Parse a standard VISA *IDN? response into an ID dict.

        :return: A dict containing the SDK version (client) and QOP version (server).
        """
        try:
            return self.qmm.version_dict()
        except (Exception,):
            return self.qmm.version()

    def set_config(self, config):
        """
        Update the configuration used by the OPX.

        :param config: the new configuration.
        """
        self.config = config

    def open_qm(self, close_other_machines: bool):
        """
        Open a quantum machine with a given configuration ready to execute a program.
        Beware that each call will close the existing quantum machines and interrupt the running jobs.
        """
        self.qm = self.qmm.open_qm(self.config, close_other_machines=close_other_machines)
        self.qm_id = self.qm.id

    def update_qm(self):
        """
        Close and re-open a new quantum machine so that it reloads the configuration in case it has been updated.
        """
        if self.qm_id in self.qmm.list_open_quantum_machines():
            self.qm.close()
        self.open_qm(self.close_other_machines)

    # Empty method that can be replaced by your pulse sequence in the main script
    # This can also be modified so that you can put the sequences here directly...
    def qua_program(self):
        """
        Custom QUA program

        :return: QUA program
        """
        with program() as prog:
            pass
        return prog

    # @abstractmethod
    def get_prog(self):
        """Get the QUA program from the user"""
        prog = self.qua_program
        return prog

    # @abstractmethod
    def get_res(self):
        """
        Fetch the results from the OPX, convert it into Volts and cast it into a dict.

        :return: dict containing the fetched results.
        """

        if self.result_handles is None:
            return None
        else:
            output = {}
            for i in range(len(self.results["types"])):
                # Get data and convert to Volt
                out = None
                # demodulated or integrated data
                self.result_handles.get(self.results["names"][i]).wait_for_values(self.counter)
                if self.results["types"][i] == "IQ":
                    out = (
                        -(self.result_handles.get(self.results["names"][i]).fetch(self.counter - 1)["value"])
                        * 4096
                        / (self.readout_pulse_length() * self.readout_sampling_rate())
                        * self.demod_factor
                        * self.results["scale_factor"][i]
                    )
                # raw adc traces
                elif self.results["types"][i] == "adc":
                    out = (
                        -(self.result_handles.get(self.results["names"][i]).fetch(self.counter - 1)["value"])
                        / 4096
                        * self.results["scale_factor"][i]
                    )
                # Reshape data
                if len(self.results["buffers"][i]) == 2:
                    output[self.results["names"][i]] = out.reshape(
                        self.results["buffers"][i][0], self.results["buffers"][i][1]
                    )
                elif len(self.results["buffers"][i]) == 1:
                    output[self.results["names"][i]] = out.reshape(self.results["buffers"][i][0])
                else:
                    output[self.results["names"][i]] = out

            # Add amplitude and phase if I and Q are in the SP
            if "I" in output.keys() and "Q" in output.keys():
                output["R"] = np.sqrt(output["I"] ** 2 + output["Q"] ** 2)
                if self.unwrap_phase is True:
                    output["Phi"] = np.unwrap(np.angle(output["I"] + 1j * output["Q"])) * 180 / np.pi
                else:
                    output["Phi"] = np.angle(output["I"] + 1j * output["Q"]) * 180 / np.pi
            return output

    def _extend_result(self, gene, count, averaging_buffer):
        """
        Recursive function to get relevant information from the stream processing to construct the result Parameter.

        :param gene: stream processing generator (item of prog.result_analysis._result_analysis.model).
        :param count: counter to keep track of the buffers for a given stream.
        :param averaging_buffer: flag identifying if a buffer is used for averaging.
        """

        if len(gene.values) > 0:
            if gene.values[0].string_value == "saveAll":
                self.results["names"].append(gene.values[1].string_value)
                self.results["types"].append("IQ")
                self.results["units"].append("V")
                self.results["buffers"].append([])
                self.results["scale_factor"].append(1)
                # Check if next buffer is for averaging
                if hasattr(gene.values[2].list_value.values[1], "list_value"):
                    if len(gene.values[2].list_value.values[1].list_value.values) > 0:
                        if gene.values[2].list_value.values[1].list_value.values[0].string_value == "average":
                            averaging_buffer = True
            elif gene.values[0].string_value == "buffer":
                if not averaging_buffer:
                    self.results["buffers"][count].append(int(gene.values[1].string_value))
                    # Check if next buffer is for averaging
                    if hasattr(gene.values[2].list_value.values[1], "list_value"):
                        if len(gene.values[2].list_value.values[1].list_value.values) > 0:
                            if gene.values[2].list_value.values[1].list_value.values[0].string_value == "average":
                                averaging_buffer = True
                else:
                    averaging_buffer = False
            elif gene.values[0].string_value == "@macro_adc_trace":
                self.results["buffers"][count].append(int(self.readout_pulse_length() * self.readout_sampling_rate()))
                self.results["types"][count] = "adc"
                self.results["scale_factor"].append(1)
            else:
                pass
            if len(gene.values) > 2:
                if hasattr(gene.values[2], "list_value"):
                    self._extend_result(gene.values[2].list_value, count, averaging_buffer)
            elif gene.values[0].string_value == "average":
                if hasattr(gene.values[1], "list_value"):
                    self._extend_result(gene.values[1].list_value, count, averaging_buffer)

    def _get_stream_processing(self, prog):
        """
        Get relevant information from the stream processing to construct the result Parameter.

        :param prog: QUA program.
        """
        count = 0
        for i in prog.result_analysis._result_analysis.model:
            self._extend_result(i, count, False)
            count += 1

    def set_sweep_parameters(self, scanned_axis, setpoints, unit=None, label=None):
        """
        Set the setpoint parameters.

        :param scanned_axis: Can be axis1 for the most inner loop and axis2 for the outer one. 3 dimensional scans or higher are not implemented.
        :param setpoints: Values of the sweep parameter as a python list.
        :param unit: Unit of the setpoint ("V" for instance).
        :param label: Label of the setpoint ("Bias voltage" for instance).
        """
        if scanned_axis == "axis1":
            self.axis1_full_list(setpoints)
            self.axis1_start(setpoints[0])
            self.axis1_stop(setpoints[-1])
            if len(setpoints) > 1:
                self.axis1_step(setpoints[1] - setpoints[0])
            else:
                self.axis1_step(0)
            self.axis1_npoints(len(setpoints))
            self.axis1_axis.unit = unit
            self.axis1_axis.label = label
        elif scanned_axis == "axis2":
            self.axis2_full_list(setpoints)
            self.axis2_start(setpoints[0])
            self.axis2_stop(setpoints[-1])
            if len(setpoints) > 1:
                self.axis2_step(setpoints[1] - setpoints[0])
            else:
                self.axis2_step(0)
            self.axis2_npoints(len(setpoints))
            self.axis2_axis.unit = unit
            self.axis2_axis.label = label

    def get_measurement_parameter(self, scale_factor=((),)):
        """
        Find the correct Parameter shape based on the stream-processing and return the measurement Parameter.

        :param scale_factor: list of tuples containing the parameter to rescale, the scale factor with respect to Volts and the new unit as in scale_factor=[(I, 0.152, "pA"), (Q, 0.152, "pA")].
        :return: Qcodes measurement parameters.
        """

        # Reset the results in case the stream processing was changed between two iterations
        self.results = {
            "names": [],
            "types": [],
            "buffers": [],
            "units": [],
            "scale_factor": [],
        }
        # Add amplitude and phase if I and Q are in the SP
        if len(self.results["names"]) == 0:
            self._get_stream_processing(self.get_prog())

            if "I" in self.results["names"] and "Q" in self.results["names"]:
                self.results["names"].append("R")
                self.results["names"].append("Phi")
                self.results["units"].append("V")
                self.results["units"].append("deg")
            if "adc" in self.results["types"]:
                self.axis1_start(0)
                self.axis1_stop(int(self.readout_pulse_length()))
                self.axis1_step(1 / self.readout_sampling_rate())
                self.axis1_npoints(int(self.readout_pulse_length() * self.readout_sampling_rate()))
                self.axis1_full_list(np.arange(self.axis1_start(), self.axis1_stop(), self.axis1_step()))
                self.axis1_axis.unit = "ns"
                self.axis1_axis.label = "Readout time"
        # Rescale the results if a scale factor is provided
        if len(scale_factor) > 0:
            if len(scale_factor[0]) > 0:
                if len(scale_factor[0]) == 3:
                    for param in scale_factor:
                        if param[0] in self.results["names"]:
                            self.results["units"][self.results["names"].index(param[0])] = param[2]
                            self.results["scale_factor"][self.results["names"].index(param[0])] = param[1]
                else:
                    raise ValueError(
                        "scale_factor must be a list of tuples with 3 elements (the result name, the scale factor and the new unit), as in [('I', 0.152, 'pA'), ]."
                    )
        if len(self.results["buffers"]) > 0 and len(self.results["buffers"][0]) > 0:
            if len(self.results["buffers"][0]) == 2:
                return ResultParameters(
                    self,
                    self.results["names"],
                    "OPX_results",
                    names=self.results["names"],
                    units=self.results["units"],
                    shapes=((self.results["buffers"][0][0], self.results["buffers"][0][1]),)
                    * len(self.results["names"]),
                    setpoints=((self.axis2_axis(), self.axis1_axis()),) * len(self.results["names"]),
                    setpoint_units=((self.axis2_axis.unit, self.axis1_axis.unit),) * len(self.results["names"]),
                    setpoint_labels=((self.axis2_axis.label, self.axis1_axis.label),) * len(self.results["names"]),
                    setpoint_names=(
                        (
                            self.axis2_axis.label.replace(" ", "").lower(),
                            self.axis1_axis.label.replace(" ", "").lower(),
                        ),
                    )
                    * len(self.results["names"]),
                )
            elif len(self.results["buffers"][0]) == 1:
                return ResultParameters(
                    self,
                    self.results["names"],
                    "OPX_results",
                    names=self.results["names"],
                    units=self.results["units"],
                    shapes=((self.results["buffers"][0][0],),) * len(self.results["names"]),
                    setpoints=((self.axis1_axis(),),) * len(self.results["names"]),
                    setpoint_units=((self.axis1_axis.unit,),) * len(self.results["names"]),
                    setpoint_labels=((self.axis1_axis.label,),) * len(self.results["names"]),
                    setpoint_names=((self.axis1_axis.label.replace(" ", "").lower(),),) * len(self.results["names"]),
                )
        else:
            return ResultParameters(
                self,
                self.results["names"],
                "OPX_results",
                names=self.results["names"],
                units=self.results["units"],
                shapes=((),) * len(self.results["names"]),
                setpoints=((),) * len(self.results["names"]),
                setpoint_units=((),) * len(self.results["names"]),
                setpoint_labels=((),) * len(self.results["names"]),
            )

    def update_readout_length(self, readout_element: str, readout_operation: str, new_length: int):
        """
        Update the readout length of a given readout operation and readout element.
        This only works if the corresponding integration weights are constant.

        **Warning**: this function only updates the config in the current environment.
        The configuration.py file needs to be modified manually if one wishes to permanently update the readout length.

        :param readout_element: the readout element to update.
        :param readout_operation: the operation to update.
        :param new_length: the new readout length in ns - Must be a multiple of 4ns and larger than 16ns.
        """
        assert new_length % 4 == 0, "The readout length must be a multiple of 4ns."
        assert new_length > 15, "The minimum readout length is 16ns."

        config = self.config
        pulse = config["elements"][readout_element]["operations"][readout_operation]

        # Update length
        config["pulses"][pulse]["length"] = new_length
        # Update integration weights
        for weight in config["pulses"][pulse]["integration_weights"].values():
            iw = config["integration_weights"][weight]
            if len(iw["cosine"]) == 1 and len(iw["sine"]) == 1:
                value_cos = iw["cosine"][0][0]
                value_sin = iw["sine"][0][0]
                iw["cosine"] = [(value_cos, new_length)]
                iw["sine"] = [(value_sin, new_length)]
            else:
                raise RuntimeError(
                    "This method the update the readout length only works if the corresponding integration weights are constant."
                )
        # Update the quantum machine
        self.update_qm()
        print(
            f"The duration of the operation '{readout_operation}' from element '{readout_element}' is now {new_length} ns"
        )

    def live_plotting(self, results_to_plot: list = (), number_of_runs: int = 0):
        """
        Fetch and plot the specified OPX results while the program is running.

        **Warning:** This method will only work if used when no external parameters are being swept (no dond),
        because it requires the averaging to be done on the most outer loop and with the *.average()* method in the
        stream_processing as opposed to *.buffer(n_avg).map(FUNCTIONS.average())*.

        :param results_to_plot: list of the streamed data to be plotted in real-time.
        :param number_of_runs: Total number of averaging loops.
        """
        # Get the plotting grid
        if len(results_to_plot) == 0:
            raise ValueError("At least 1 result to plot must be provided")
        elif len(results_to_plot) == 2:
            grid = 1
        elif len(results_to_plot) == 3:
            grid = 121
        elif len(results_to_plot) == 4 or len(results_to_plot) == 5:
            grid = 221
        else:
            raise ValueError("Live plotting is limited to 4 parameters.")
        # Get results from QUA program
        results = fetching_tool(self.job, results_to_plot, "live")
        progress = 0
        # Live plotting
        fig = plt.figure()
        interrupt_on_close(fig, self.job)  # Interrupts the job when closing the figure
        while results.is_processing():
            # Fetch results
            data_list = results.fetch_all()
            # Subplot counter
            i = 0
            for data in data_list:
                if results_to_plot[i] == "iteration":
                    progress = data + 1
                    if data + 1 == number_of_runs:
                        self.job.halt()

                else:
                    # Convert the results into Volts
                    data = (
                        -data[-1]
                        * 4096
                        / int(self.readout_pulse_length() * self.readout_sampling_rate())
                        * self.demod_factor
                    )
                    # Plot results
                    if len(data.shape) == 1:
                        if len(results_to_plot) > 1:
                            plt.subplot(grid + i)
                        plt.cla()
                        plt.plot(self.axis1_axis(), data)
                        plt.xlabel(self.axis1_axis.label + f" [{self.axis1_axis.unit}]")
                        plt.ylabel(results_to_plot[i] + " [V]")
                    elif len(data.shape) == 2:
                        if len(results_to_plot) > 1:
                            plt.subplot(grid + i)
                        plt.cla()
                        plt.pcolor(self.axis1_axis(), self.axis2_axis(), data)
                        plt.xlabel(self.axis1_axis.label + f" [{self.axis1_axis.unit}]")
                        plt.ylabel(self.axis2_axis.label + f" [{self.axis2_axis.unit}]")
                        plt.title(results_to_plot[i] + " [V]")
                i += 1

            plt.suptitle(f"Iteration: {progress}/{number_of_runs} = {progress / number_of_runs * 100:.1f} %")
            plt.pause(1)
            plt.tight_layout()

    def run_exp(self):
        """
        Execute a given QUA program, initialize the counter to 0 and creates a result handle to fetch the results.
        """
        prog = self.get_prog()
        if " demod" in generate_qua_script(prog, self.config):
            self.demod_factor = 2
        else:
            self.demod_factor = 1
        self.job = self.qm.execute(prog)
        self.counter = 0
        self.result_handles = self.job.result_handles

    def resume(self, timeout: int = 30):
        """
        Resume the job and increment the counter to keep track of the fetched results.

        :param timeout: duration in seconds after which the console will be freed even if the pause statement has not been reached to prevent from being stuck here forever.
        """
        wait_until_job_is_paused(self.job, timeout)
        if not self.job.is_paused():
            raise RuntimeError(f"The program has not reached the pause statement before {timeout} s.")
        else:
            self.job.resume()
            self.counter += 1

    def compile_prog(self, prog):
        """
        Compile a given QUA program and stores it under the prog_id attribute.

        :param prog: A QUA program to be compiled.
        """
        self.prog_id = self.qm.compile(prog)

    def execute_compiled_prog(self):
        """
        Add a compiled program to the current queue and create a result handle to fetch the results.
        """
        if self.prog_id is not None:
            pending_job = self.qm.queue.add_compiled(self.prog_id)
            self.job = pending_job.wait_for_execution()
            self.result_handles = self.job.result_handles

    def simulate(self):
        """
        Simulate a given QUA program and store the simulated waveform into the simulated_wf attribute.
        """
        prog = self.get_prog()
        self.job = self.qmm.simulate(self.config, prog, SimulationConfig(self.sim_time() // 4))
        simulated_samples = self.job.get_simulated_samples()
        for con in simulated_samples.keys():
            self.simulated_wf[con] = {}
            self.simulated_wf[con]["analog"] = self.job.get_simulated_samples().__dict__[con].analog
            self.simulated_wf[con]["digital"] = self.job.get_simulated_samples().__dict__[con].digital
        self.result_handles = self.job.result_handles

    def plot_simulated_wf(self):
        """
        Plot the simulated waveforms in a new figure using matplotlib.
        """
        plt.figure()
        for con in self.simulated_wf.keys():
            for t in self.simulated_wf[con].keys():
                for port in self.simulated_wf[con][t].keys():
                    if not np.all(self.simulated_wf[con][t][port] == 0):
                        if len(self.simulated_wf.keys()) == 1:
                            plt.plot(self.simulated_wf[con][t][port], label=f"{t} {port}")
                        else:
                            plt.plot(
                                self.simulated_wf[con][t][port],
                                label=f"{con} {t} {port}",
                            )
        plt.xlabel("Time [ns]")
        plt.ylabel("Voltage level [V]")
        plt.title("Simulated waveforms")
        plt.grid("on")
        plt.legend()

    def close(self) -> None:
        """
        Close the quantum machine and tear down the OPX instrument.
        """
        if self.qm_id in self.qmm.list_open_quantum_machines():
            self.qm.close()
        super().close()

    def halt(self) -> None:
        """
        Interrupt the current job and halt the running program.
        """
        if self.job is not None:
            self.job.halt()


# noinspection PyAbstractClass
class ResultParameters(MultiParameter):
    """
    Subclass of MultiParameter suited for the results acquired with the OPX.
    """

    def __init__(
        self,
        instr,
        params,
        name,
        names,
        units,
        shapes=None,
        setpoints=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            name=name,
            names=names,
            units=units,
            shapes=shapes,
            setpoints=setpoints,
            *args,
            **kwargs,
        )

        self._instr = instr
        self._params = params

    def get_raw(self):
        vals = []
        result = self._instr.get_res()
        for param in self._params:
            if param in self._instr.results["names"]:
                vals.append(result[param])
        return tuple(vals)


# noinspection PyAbstractClass
class GeneratedSetPoints(Parameter):
    """
    A parameter that generates a setpoint array from start, stop and num points
    parameters.
    """

    def __init__(self, startparam, stopparam, numpointsparam, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._startparam = startparam
        self._stopparam = stopparam
        self._numpointsparam = numpointsparam

    def get_raw(self):
        return np.linspace(self._startparam(), self._stopparam(), self._numpointsparam())


# noinspection PyAbstractClass
class GeneratedSetPointsSpan(Parameter):
    """
    A parameter that generates a setpoint array from center, span and num points
    parameters.
    """

    def __init__(self, spanparam, centerparam, numpointsparam, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._spanparam = spanparam
        self._centerparam = centerparam
        self._numpointsparam = numpointsparam

    def get_raw(self):
        return np.linspace(
            self._centerparam() - self._spanparam() / 2,
            self._centerparam() + self._spanparam() / 2,
            self._numpointsparam(),
        )


# noinspection PyAbstractClass
class GeneratedSetPointsArbitrary(Parameter):
    """
    A parameter that generates a setpoint array from an arbitrary list of points.
    """

    def __init__(self, full_list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.full_list = full_list

    def get_raw(self):
        return self.full_list()



# Map InstrumentChannel.instrument_id to containing instrument type
instrument_id_mapping = {
    "lf-fem": "OPX1000",
    "mw-fem": "OPX1000",
    "opx+": "OPX+",
    "octave": "Octave",
    "external-mixer": "Mixers",
}

# Define the chassis dimensions
INSTRUMENT_FIGURE_DIMENSIONS = {
    "OPX1000": {"width": 8, "height": 3},
    "OPX+": {"width": 8, "height": 1},
    "Octave": {"width": 3, "height": 1},
    "Mixers": {"width": 1, "height": 1},
}

OPX_PLUS_ASPECT = INSTRUMENT_FIGURE_DIMENSIONS["OPX+"]["height"] / INSTRUMENT_FIGURE_DIMENSIONS["OPX+"]["width"]
OPX_1000_ASPECT = INSTRUMENT_FIGURE_DIMENSIONS["OPX1000"]["height"] / INSTRUMENT_FIGURE_DIMENSIONS["OPX1000"]["width"]
# Define the port positions for different modules
PORT_SPACING_FACTOR = 0.12
PORT_SIZE = 0.055

fem_analog_output_positions = [(0.05 + 0.25 * OPX_1000_ASPECT, 1.06 - i * PORT_SPACING_FACTOR) for i in range(8)]
fem_digital_output_positions = [(0.05 + 0.75 * OPX_1000_ASPECT, 0.86 - i * PORT_SPACING_FACTOR / 2.4) for i in range(8)]

PORT_POSITIONS = {
    "lf-fem": {
        "analog": {
            "output": fem_analog_output_positions,
            "input": [(0.05 + 0.75 * OPX_1000_ASPECT, 1.06 - (6 + i) * PORT_SPACING_FACTOR) for i in range(2)],
        },
        "digital": {"output": fem_digital_output_positions},
    },
    "mw-fem": {
        "analog": {
            "output": fem_analog_output_positions,
            "input": [(0.05 + 0.75 * OPX_1000_ASPECT, 1.06 - i * PORT_SPACING_FACTOR * 7) for i in range(2)],
        },
        "digital": {"output": fem_digital_output_positions},
    },
    "opx+": {
        "analog": {
            "input": [(1.1 * 3, 0.01 + 0.32 / (i + 1)) for i in range(2)],
            "output": [((0.7 + j * 0.06) * 3, 0.01 + 0.32 / (i + 1)) for j in range(5) for i in range(2)],
        },
        "digital": {
            "output": [((0.3 + j * 0.06) * 3, 0.01 + 0.32 / (i + 1)) for j in range(5) for i in range(2)],
        },
    },
    "octave": {
        "analog": {
            "input": [((0.19 + j * 0.06) * 3, 0.25) for j in range(2)],
            "output": [((0.3 + j * 0.06) * 3, 0.32) for j in range(5)],
            "inter_input": [(1.1 * 3, 0.01 + 0.32 / (i + 1)) for i in range(2)],
            "inter_output": [((0.7 + j * 0.06) * 3, 0.01 + 0.32 / (i + 1)) for j in range(5) for i in range(2)],
        },
        "digital": {
            "input": [((0.3 + j * 0.06) * 3, 0.18) for j in range(5)],
        },
    },
    "external-mixer": {
        "analog": {
            "input": [(0.75, 0.45)],
            "output": [(0.25, 0.45)],
        },
        "digital": {
            "input": [(0.25, 0.85)],
        },
    },
}



from qualang_tools.digital_filters.filters import (
    QOPVersion,
    calc_filter_taps,
    exponential_decay,
    high_pass_exponential,
    single_exponential_correction,
    multi_exponential_decay,
    highpass_correction,
    bounce_and_delay_correction,
)

__all__ = [
    "QOPVersion",
    "calc_filter_taps",
    "exponential_decay",
    "high_pass_exponential",
    "single_exponential_correction",
    "multi_exponential_decay",
    "highpass_correction",
    "bounce_and_delay_correction",
]



NUM_OCTAVE_INPUT_PORTS = 2
NUM_OCTAVE_DIGITAL_INPUT_PORTS = 5
NUM_OCTAVE_OUTPUT_PORTS = 5
NUM_OPX_PLUS_INPUT_PORTS = 2
NUM_OPX_PLUS_OUTPUT_PORTS = 10
NUM_OPX_PLUS_DIGITAL_OUTPUT_PORTS = 10
NUM_LF_FEM_INPUT_PORTS = 2
NUM_LF_FEM_OUTPUT_PORTS = 8
NUM_LF_FEM_DIGITAL_OUTPUT_PORTS = 8
NUM_MW_FEM_INPUT_PORTS = 2
NUM_MW_FEM_OUTPUT_PORTS = 8
NUM_MW_FEM_DIGITAL_OUTPUT_PORTS = 8



"""Tools to help handling results from QUA programs.

Content:
    - fetching_tool: API to easily fetch data from the stream processing.
    - progress_counter: Displays progress bar and prints remaining computation time.
"""

import numpy as np
import time
from qm.jobs.running_qm_job import RunningQmJob
from warnings import warn


class fetching_tool:
    def __init__(self, job, data_list, mode="wait_for_all"):
        """
        API to easily fetch data from the stream processing.
        **Example**: my_results = fetching_tool(job=my_job, data_list=["I", "Q"], mode="live")

        :param job: a ``QmJob`` object (see QM Job API) corresponding to the current execution.
        :param data_list: a list of the result names saved in the stream processing.
        :param mode: current acquisition mode among ['live', 'wait_for_all'], default is 'wait_for_all'. 'live' will fetch data one by one for all named results for live plotting purposes. 'wait_for_all' will wait until all values were processed for all named results.
        """
        if not data_list:
            raise Exception("The provided data list is empty.")
        if mode not in ["live", "wait_for_all"]:
            raise Exception(f"Mode '{mode}' is not supported. Supported modes are ['live', 'wait_for_all']")
        self.data_list = data_list
        self.mode = mode
        self.results = []
        self.data_handles = []
        self.res_handles = job.result_handles
        if mode == "live":
            for data in self.data_list:
                if hasattr(self.res_handles, data):
                    self.data_handles.append(self.res_handles.get(data))
                    self.data_handles[-1].wait_for_values(1)
                else:
                    raise Warning(f"{data} is not saved in the stream processing.")
            # Live plotting parameters
            self._b_cont = False
            self._b_last = True
            self.start_time = 0

    def is_processing(self):
        """
        Returns True while the program is processing, and also once after the processing is done. Can be used for live plotting.
        **Example**: while my_results.is_processing():

        :return: boolean flag which is True while the program is processing, and also once after the processing is done.
        """
        if self.start_time == 0:
            self._b_cont = self.res_handles.is_processing()
            self._b_last = not self._b_cont
            self.start_time = time.time()
        else:
            self._b_cont = self.res_handles.is_processing()
            self._b_last = not (self._b_cont or self._b_last)
        return self._b_cont or self._b_last

    def _format(self, data):
        if type(data) is np.ndarray:
            if type(data[0]) is np.void:
                if len(data.dtype) == 1:
                    data = data["value"]
        return data

    def get_start_time(self):
        """
        Gets time at which is_processing() was first called. To be used in progress_counter().

        :return: float for the time at which is_processing() was first called.
        """
        return self.start_time

    def fetch_all(self):
        """
        Fetch a result from the current result stream saved in server memory. The result stream is populated by the
        save() and save_all() statements. Note that if timestamps are saved with the `with_timestamp()` method, then
        the result will contain both the values and timestamps.
        **Example**: I, Q = my_results.fetch_all()

        :return: all result of current result stream as a list of python variables
        """
        if self.mode == "wait_for_all":
            self.res_handles.wait_for_all_values()
            for data in self.data_list:
                if hasattr(self.res_handles, data):
                    self.results.append(self._format(self.res_handles.get(data).fetch_all()))

        elif self.mode == "live":
            self.results = []
            for i in range(len(self.data_handles)):
                self.results.append(self._format(self.data_handles[i].fetch_all()))
        return self.results


def progress_counter(iteration, total, progress_bar=True, percent=True, start_time=None):
    """Displays progress bar and prints remaining computation time.

    :param iteration: current iteration. Must be a python integer.
    :param total: total number of iteration. Must be a python integer.
    :param progress_bar: Flag enabling the progress bar display. Must be True or False. Default is True.
    :param percent: Flag enabling the progress percentage display. Must be True or False. Default is True.
    :param start_time: Starting time of the program enabling the elapsed time display. Must be the result of time.time(). Default is None.
    :return: None.
    """
    current_percent = (iteration + 1) / total * 100

    progress = "Progress: "
    if progress_bar:
        bar = "#" * int(current_percent / 2)
        progress += "[" + bar + " " * (50 - len(bar)) + "] "
    if percent:
        progress += f"{current_percent:.1f}% (n={iteration + 1}/{total})"
    if start_time is not None:
        progress += f" --> elapsed time: {time.time() - start_time:.2f}s"

    print(progress, end="\r")
    if current_percent == 100:
        print("")


def wait_until_job_is_paused(running_job: RunningQmJob, timeout: int = 30, strict_timeout: bool = True):
    """
    Waits until the OPX FPGA reaches a "pause" statement.
    Used when the OPX sequence needs to be synchronized with an external parameter sweep and to ensure that the OPX
    sequence is done before proceeding to the next iteration of the external loop, or when using IO variables:
    https://docs.quantum-machines.co/0.1/qm-qua-sdk/docs/Guides/features/?h=pause#pause-resume-and-io-variables

    :param running_job: the QM running job object.
    :param timeout: duration in seconds after which the console will be freed even if the pause statement has not been reached to prevent from being stuck here forever.
    :param strict_timeout: will throw and exception is set to True, otherwise it will just a print a warning.
    :return: True when the pause statement has been reached.
    """
    start = time.time()
    delay = 0
    while (not running_job.is_paused()) and (delay < timeout):
        time.sleep(0.1)
        delay = time.time() - start
    if delay > timeout:
        if strict_timeout:
            raise TimeoutError(f"Timeout ({timeout}s) was reached, consider extending it if it was not intended.")
        else:
            warn(f"Timeout ({timeout}s) was reached, consider extending it if it was not intended.")
    return True



from collections import Counter

from qualang_tools.wirer.connectivity.wiring_spec import WiringSpec
from qualang_tools.wirer.instruments.instrument_channel import AnyInstrumentChannel


class ChannelSpec:
    """A specification for fixed channel requirements."""

    def __init__(self):
        self.channel_templates = None

    def __and__(self, other: "ChannelSpec") -> "ChannelSpec":
        combined_channel_spec = ChannelSpec()
        combined_channel_spec.channel_templates = self.channel_templates + other.channel_templates
        return combined_channel_spec

    def is_empty(self):
        return self.channel_templates is None or self.channel_templates == []

    def filter_by_wiring_spec(self, wiring_spec: WiringSpec) -> "ChannelSpec":
        """
        Filters out channels from the specification if:
            1. Their I/O type is not required by the wiring specification
            2. They are digital but the wiring specification doesn't require triggering
        """

        def filter_func(channel: AnyInstrumentChannel):
            analog_channel_matches_io_type = (
                channel.signal_type == "digital" or channel.io_type in wiring_spec.io_type.value
            )
            digital_channel_only_if_triggered = channel.signal_type == "analog" or wiring_spec.triggered
            return analog_channel_matches_io_type and digital_channel_only_if_triggered

        filtered_channel_spec = ChannelSpec()
        filtered_channel_spec.channel_templates = list(filter(filter_func, self.channel_templates))

        return filtered_channel_spec

    def apply_constraints(self, constraints: "ChannelSpec") -> bool:
        """
        Attempt to constrain the channel specifications according to the
        filled attributes of all the specs in the `constraints` list.

        Returns:
            True: if constraints are successful, modifying the spec.
            False: if any of the constraints cannot be applied and are not already met.
        """
        # make sure there are at least enough channel templates in this object
        # of the correct type to satisfy each of the provided constraints.
        # e.g., it is invalid to apply a MW-FEM mask if there are only OPX+
        # channel templates present in this object.
        counts_templates = Counter([type(spec) for spec in self.channel_templates])
        counts_constraints = Counter([type(spec) for spec in constraints.channel_templates])

        for _type, count_constraints in counts_constraints.items():
            if count_constraints > counts_templates.get(_type, 0):
                return False

        # try to constraints to any possible spec in the list in order that
        # the constraint is provided.
        constrained_templates = []
        for constraint in constraints.channel_templates:
            constrained = False
            for template in self.channel_templates:
                if template in constrained_templates:
                    # can't constrain already constrained spec
                    continue
                if type(template) is not type(constraint):
                    # can't constrain a spec of a different type
                    continue
                else:
                    attrs_to_constrain = ["con", "slot", "port"]
                    attrs_constrainable = 0
                    # make sure every filled attribute of the constraint can be
                    # applied to the spec without interfering with existing
                    # constraints.
                    for attr in attrs_to_constrain:
                        spec_attr = getattr(template, attr, None)
                        constraint_attr = getattr(constraint, attr, None)

                        constrainable = spec_attr is None
                        nothing_to_constrain = constraint_attr is None
                        constrained_already_met = constraint_attr == spec_attr

                        if constrainable or nothing_to_constrain or constrained_already_met:
                            attrs_constrainable += 1
                        else:
                            break

                    if attrs_constrainable == len(attrs_to_constrain):
                        for attr in attrs_to_constrain:
                            setattr(template, attr, getattr(constraint, attr, None))
                        constrained = True
                        constrained_templates.append(template)
                        break

            if not constrained:
                return False

        return True



from pathlib import Path
from abc import ABC
from typing import Dict, Any, Generator, List, Tuple, Optional
from matplotlib import pyplot as plt
import numpy as np

__all__ = ["DEFAULT_DATA_PROCESSORS", "DataProcessor", "MatplotlibPlotSaver", "NumpyArraySaver", "XarraySaver"]


DEFAULT_DATA_PROCESSORS = []


def iterate_nested_dict(
    d: Dict[str, Any], parent_keys: Optional[List[str]] = None
) -> Generator[Tuple[List[str], Any], None, None]:
    """Iterate over a nested dictionary

    :param d: The dictionary to iterate over
    :param parent_keys: The keys of the parent dictionary. Used for recursion

    :return: A generator that yields a tuple of the keys and the value

    """
    if parent_keys is None:
        parent_keys = []
    for k, v in d.items():
        keys = parent_keys + [k]
        yield keys, v
        if isinstance(v, dict):
            yield from iterate_nested_dict(v, parent_keys=keys)


def update_nested_dict(d: dict, keys: List[Any], value: Any) -> None:
    """Update a nested dictionary with a new value

    :param d: The dictionary to update
    :param keys: The keys to the value to update
    :param value: The new value to set
    """
    subdict = d
    for key in keys[:-1]:
        subdict = subdict[key]

    subdict[keys[-1]] = value


def copy_nested_dict(d: dict) -> dict:
    """Copy a nested dictionary, but don't make copies of the values

    This function will copy a nested dictionary, but will not make copies of the values. This is useful if copying the
    values may be an expensive operation (e.g. large arrays).
    If you also want to make copies of the values, use `copy.deepcopy`

    :param d: The dictionary to copy
    :return: A new dictionary with the same structure as `d`, but with the same values
    """
    new_dict = {}
    for key, val in d.items():
        if isinstance(val, dict):
            new_dict[key] = copy_nested_dict(val)
        else:
            new_dict[key] = val
    return new_dict


class DataProcessor(ABC):
    def process(self, data):
        return data

    def post_process(self, data_folder: Path):
        pass


class MatplotlibPlotSaver(DataProcessor):
    file_format: str = "png"
    nested_separator: str = "."

    def __init__(self, file_format=None):
        if file_format is not None:
            self.file_format = file_format
        self.data_figures = {}

    @property
    def file_suffix(self):
        suffix = self.file_format
        if not suffix.startswith("."):
            suffix = "." + suffix
        return suffix

    def process(self, data):
        self.data_figures = {}

        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, plt.Figure):
                continue

            file_end = Path(keys[-1]).with_suffix(self.file_suffix)
            path = Path(self.nested_separator.join(keys[:-1] + [str(file_end)]))

            self.data_figures[path] = val
            update_nested_dict(processed_data, keys, f"./{path}")

        return processed_data

    def post_process(self, data_folder: Path):
        for path, fig in self.data_figures.items():
            fig.savefig(data_folder / path, bbox_inches="tight")


DEFAULT_DATA_PROCESSORS.append(MatplotlibPlotSaver)


class NumpyArraySaver(DataProcessor):
    merge_arrays: bool = True
    merged_array_name: str = "arrays.npz"
    nested_separator: str = "."

    def __init__(self, merge_arrays=None, merged_array_name=None):
        if merge_arrays is not None:
            self.merge_arrays = merge_arrays
        if merged_array_name is not None:
            self.merged_array_name = merged_array_name

        self.data_arrays = {}

    def process(self, data):
        self.data_arrays = {}
        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, np.ndarray):
                continue

            path = Path(self.nested_separator.join(keys))
            self.data_arrays[path] = val
            if self.merge_arrays:
                update_nested_dict(processed_data, keys, f"./{self.merged_array_name}#{path}")
            else:
                update_nested_dict(processed_data, keys, f"./{path}.npy")
        return processed_data

    def post_process(self, data_folder: Path):
        if self.merge_arrays:
            arrays = {str(path): arr for path, arr in self.data_arrays.items()}
            if arrays:
                np.savez(data_folder / self.merged_array_name, **arrays)
        else:
            for path, arr in self.data_arrays.items():
                np.save(data_folder / f"{path}.npy", arr)


DEFAULT_DATA_PROCESSORS.append(NumpyArraySaver)


class XarraySaver(DataProcessor):
    merge_arrays: bool = False
    merged_array_name: str = "xarrays"
    file_format: str = "hdf5"
    nested_separator: str = "."

    def __init__(self, merge_arrays=None, merged_array_name=None, file_format=None):
        if merge_arrays is not None:
            self.merge_arrays = merge_arrays
        if merged_array_name is not None:
            self.merged_array_name = merged_array_name
        if file_format is not None:
            self.file_format = file_format

        self.data_arrays = {}

    @property
    def file_suffix(self) -> str:
        suffixes = {"nc": ".nc", "netcdf": ".nc", "h5": ".h5", "hdf5": ".h5", "zarr": ".zarr"}
        return suffixes[self.file_format.lower()]

    def process(self, data):
        import xarray as xr

        self.data_arrays = {}
        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, xr.Dataset):
                continue

            path = Path(self.nested_separator.join(keys))
            self.data_arrays[path] = val
            if self.merge_arrays:
                merged_array_name = Path(self.merged_array_name).with_suffix(self.file_suffix)
                update_nested_dict(processed_data, keys, f"./{merged_array_name}#{path}")
            else:
                update_nested_dict(processed_data, keys, f"./{path}{self.file_suffix}")
        return processed_data

    def save_merged_netcdf_arrays(self, path: Path, arrays: dict):
        for array_path, array in self.data_arrays.items():
            try:
                array.to_netcdf(path, mode="a", group=str(array_path))
            except ValueError as e:
                raise ValueError(
                    f"Error saving merged array {path}. You may neet to first run `pip install netcdf4`"
                ) from e

    def post_process(self, data_folder: Path):
        if self.file_suffix not in [".nc", ".h5"]:
            raise NotImplementedError(f"File format {self.file_format} is not supported")

        if self.merge_arrays:
            for path, array in self.data_arrays.items():
                merged_path = data_folder / Path(self.merged_array_name).with_suffix(self.file_suffix)
                try:
                    array.to_netcdf(merged_path, mode="a", group=str(path))
                except ValueError as e:
                    raise ValueError(
                        f"Error saving merged array {merged_path}. You may neet to first run `pip install netcdf4`"
                    ) from e
        else:
            for path, array in self.data_arrays.items():
                array.to_netcdf(data_folder / f"{path}{self.file_suffix}")


try:
    import xarray  # noqa: F401

    DEFAULT_DATA_PROCESSORS.append(XarraySaver)
except ImportError:
    pass



from qualang_tools.wirer.connectivity.channel_spec import ChannelSpec
from qualang_tools.wirer.instruments.instrument_channel import (
    InstrumentChannel,
    InstrumentChannelLfFemInput,
    InstrumentChannelLfFemOutput,
    InstrumentChannelMwFemInput,
    InstrumentChannelMwFemOutput,
    InstrumentChannelOpxPlusInput,
    InstrumentChannelOpxPlusOutput,
    InstrumentChannelOctaveInput,
    InstrumentChannelOctaveOutput,
    InstrumentChannelExternalMixerInput,
    InstrumentChannelExternalMixerOutput,
    InstrumentChannelOpxPlusDigitalOutput,
    InstrumentChannelMwFemDigitalOutput,
    InstrumentChannelLfFemDigitalOutput,
    InstrumentChannelOctaveDigitalInput,
    InstrumentChannelExternalMixerDigitalInput,
)

# A channel template is a partially filled InstrumentChannel object
ChannelTemplate = InstrumentChannel


class ChannelSpecMwFemSingle(ChannelSpec):
    def __init__(self, con: int = None, slot: int = None, in_port: int = None, out_port: int = None):

        super().__init__()
        self.channel_templates = [
            InstrumentChannelMwFemInput(con=con, slot=slot, port=in_port),
            InstrumentChannelMwFemOutput(con=con, slot=slot, port=out_port),
        ]


class ChannelSpecLfFemSingle(ChannelSpec):
    def __init__(
        self, con: int = None, in_slot: int = None, in_port: int = None, out_slot: int = None, out_port: int = None
    ):

        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=in_slot, port=in_port),
            InstrumentChannelLfFemOutput(con=con, slot=out_slot, port=out_port),
        ]


class ChannelSpecLfFemBaseband(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
    ):

        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
        ]


class ChannelSpecOpxPlusSingle(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None, out_port: int = None):

        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port),
        ]


class ChannelSpecOpxPlusBaseband(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
    ):

        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
        ]


class ChannelSpecOctave(ChannelSpec):
    def __init__(self, index: int = None, rf_in: int = None, rf_out: int = None):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOctaveInput(con=index, port=rf_in),
            InstrumentChannelOctaveOutput(con=index, port=rf_out),
        ]


class ChannelSpecExternalMixer(ChannelSpec):
    def __init__(self, index: int = None, rf_in: int = None, rf_out: int = None):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelExternalMixerInput(con=index, port=rf_in),
            InstrumentChannelExternalMixerOutput(con=index, port=rf_out),
        ]


class ChannelSpecLfFemBasebandAndOctave(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        octave_index: int = None,
        rf_in: int = None,
        rf_out: int = None,
    ):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
            InstrumentChannelOctaveInput(con=octave_index, port=rf_in),
            InstrumentChannelOctaveOutput(con=octave_index, port=rf_out),
        ]


class ChannelSpecLfFemBasebandAndExternalMixer(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        slot: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        mixer_index: int = None,
    ):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_i),
            InstrumentChannelLfFemInput(con=con, slot=slot, port=in_port_q),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_i),
            InstrumentChannelLfFemOutput(con=con, slot=slot, port=out_port_q),
            InstrumentChannelExternalMixerInput(con=mixer_index, port=1),
            InstrumentChannelExternalMixerOutput(con=mixer_index, port=1),
        ]


class ChannelSpecOpxPlusBasebandAndOctave(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        octave_index: int = None,
        rf_in: int = None,
        rf_out: int = None,
    ):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
            InstrumentChannelOctaveInput(con=octave_index, port=rf_in),
            InstrumentChannelOctaveOutput(con=octave_index, port=rf_out),
        ]


class ChannelSpecOpxPlusBasebandAndExternalMixer(ChannelSpec):
    def __init__(
        self,
        con: int = None,
        in_port_i: int = None,
        in_port_q: int = None,
        out_port_i: int = None,
        out_port_q: int = None,
        mixer_index: int = None,
    ):
        super().__init__()
        self.channel_templates = [
            InstrumentChannelOpxPlusInput(con=con, port=in_port_i),
            InstrumentChannelOpxPlusInput(con=con, port=in_port_q),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_i),
            InstrumentChannelOpxPlusOutput(con=con, port=out_port_q),
            InstrumentChannelExternalMixerInput(con=mixer_index, port=1),
            InstrumentChannelExternalMixerOutput(con=mixer_index, port=1),
        ]


class ChannelSpecOpxPlusDigital(ChannelSpec):
    def __init__(self, con: int = None, out_port: int = None):
        super().__init__()
        self.channel_templates = [InstrumentChannelOpxPlusDigitalOutput(con=con, port=out_port)]


class ChannelSpecMwFemDigital(ChannelSpec):
    def __init__(self, con: int = None, out_port: int = None):
        super().__init__()
        self.channel_templates = [InstrumentChannelMwFemDigitalOutput(con=con, port=out_port)]


class ChannelSpecLfFemDigital(ChannelSpec):
    def __init__(self, con: int = None, out_port: int = None):
        super().__init__()
        self.channel_templates = [InstrumentChannelLfFemDigitalOutput(con=con, port=out_port)]


class ChannelSpecOctaveDigital(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None):
        super().__init__()
        self.channel_templates = [InstrumentChannelOctaveDigitalInput(con=con, port=in_port)]


class ChannelSpecExternalMixerDigital(ChannelSpec):
    def __init__(self, con: int = None, in_port: int = None):
        super().__init__()
        self.channel_templates = [InstrumentChannelExternalMixerDigitalInput(con=con, port=in_port)]


mw_fem_spec = ChannelSpecMwFemSingle
lf_fem_spec = ChannelSpecLfFemSingle
lf_fem_iq_spec = ChannelSpecLfFemBaseband
lf_fem_iq_octave_spec = ChannelSpecLfFemBasebandAndOctave
lf_fem_iq_ext_mixer_spec = ChannelSpecLfFemBasebandAndExternalMixer
opx_spec = ChannelSpecOpxPlusSingle
opx_iq_spec = ChannelSpecOpxPlusBaseband
opx_iq_octave_spec = ChannelSpecOpxPlusBasebandAndOctave
opx_iq_ext_mixer_spec = ChannelSpecOpxPlusBasebandAndExternalMixer
octave_spec = ChannelSpecOctave
ext_mixer_spec = ChannelSpecExternalMixer
opx_dig_spec = ChannelSpecOpxPlusDigital



from typing import Dict, Optional
from qm.api.models.compiler import CompilerOptionArguments
from qm.jobs.running_qm_job import RunningQmJob


from qm.program import Program
from qm.qua._dsl import _ProgramScope as _ProgramScope_qua
from qm import QuantumMachine
from qm.simulate.interface import SimulationConfig


__all__ = [
    "patch_qua_program_addons",
]


class _ProgramScope(_ProgramScope_qua):
    def __enter__(self) -> Program:
        program = super().__enter__()

        # Instantiate all addons
        program.addons = {name: addon() for name, addon in Program.addons.items()}

        # TODO Should we add exception handling?
        for addon in self.program.addons.values():
            addon.enter_program(program)

        return self.program

    def __exit__(self, exc_type, exc_val, exc_tb):
        for addon in self.program.addons.values():
            # TODO Should we add exception handling?
            addon.exit_program(exc_type, exc_val, exc_tb)

        exit_result = super().__exit__(exc_type, exc_val, exc_tb)

        return exit_result


QM_execute = QuantumMachine.execute


def QM_execute_patched(
    self,
    program: Program,
    duration_limit: int = 1000,
    data_limit: int = 20000,
    force_execution: int = False,
    dry_run: int = False,
    simulate: Optional[SimulationConfig] = None,
    compiler_options: Optional[CompilerOptionArguments] = None,
) -> RunningQmJob:
    return_val = QM_execute(
        self,
        program,
        duration_limit,
        data_limit,
        force_execution,
        dry_run,
        simulate,
        compiler_options,
    )

    for program_addon in program.addons.values():
        program_addon.execute_program(program=program, quantum_machine=self)

    return return_val


def patch_qua_program_addons():
    import qm.qua._dsl

    if hasattr(Program, "addons"):
        print("qm.program.Program already has 'addons' attribute, not patching")
    else:
        from qualang_tools.callable_from_qua import ProgramAddon

        Program.addons: Dict[str, ProgramAddon] = {}

    if qm.qua._dsl._ProgramScope is _ProgramScope:
        print("qm.qua._dsl._ProgramScope already patched, not patching")
    else:
        qm.qua._dsl._ProgramScope = _ProgramScope

    if qm.QuantumMachine.execute is QM_execute_patched:
        print("qm.QuantumMachine.QuantumMachine.execute already patched, not patching")
    else:
        qm.QuantumMachine.execute = QM_execute_patched



from typing import List

from ..connectivity.channel_spec import ChannelSpec
from ..connectivity.wiring_spec import WiringSpec


class ConstraintsTooStrictException(Exception):
    def __init__(self, wiring_spec: WiringSpec, constraints: List[ChannelSpec]):
        message = (
            f"Failed to find a valid channel template for {wiring_spec.frequency} "
            f"{wiring_spec.io_type.value} channels on the "
            f"{wiring_spec.line_type.value} line for elements "
            f"{','.join([str(e.id) for e in wiring_spec.elements])} with the "
            f"following constraints: {constraints}. If you are intentionally trying to "
            f"allocate multiple lines to the same port, see documentation for the "
            f"correct approach."
        )
        super(ConstraintsTooStrictException, self).__init__(message)


class NotEnoughChannelsException(Exception):
    def __init__(self, wiring_spec: WiringSpec):
        message = (
            f"Couldn't find enough available {wiring_spec.frequency.value} "
            f"{wiring_spec.io_type.value} channels satisfying the wiring "
            f"specfication for the {wiring_spec.line_type.value} line for elements "
            f"{','.join([str(e.id) for e in wiring_spec.elements])}"
        )
        super(NotEnoughChannelsException, self).__init__(message)



from typing import Union

import numpy as np
from qm.qua import declare, wait, for_
from qualang_tools.addons.calibration.calibrations import u

MAX_WAIT = 2**31 - 1  # units of clock cycles (4ns)


def long_wait(wait_time: Union[int, np.integer], *elements: str, threshold_for_looping: int = u.to_clock_cycles(1e6)):
    """Perform a QUA wait command for longer than the current maximum wait time, 2^31-1.

    This macro unrolls into QUA for-loop of smaller waits if the maximum wait time is exceeded in `wait_time`.

    Note: `wait_time` must be a python integer, not a QUA variable.

    Args:
        wait_time (int): Desired wait time in units of clock cycles (4ns).
        *elements (str): Variable length argument list of elements to wait on.
        threshold_for_looping: Minimum wait time at which `long_wait` uses a QUA for-loop of waits (default 1ms).
         1ms is chosen as a safe default because constants larger than 1e6 in a QUA program tend to extend compilation
         times. In units of clock cycles.
    """
    if not isinstance(wait_time, (int, np.integer)):
        raise TypeError(f"Expected wait_time to be a float or an integer, got {type(wait_time)}.")

    if threshold_for_looping < 4:
        raise ValueError(f"Expected loop threshold to be greater than 4 clock cycles, got {threshold_for_looping}")

    if threshold_for_looping > MAX_WAIT:
        raise ValueError(
            f"Expected loop threshold to be less than the maximum wait time {MAX_WAIT}, got {threshold_for_looping}"
        )

    if wait_time < threshold_for_looping:
        wait(wait_time, *elements)

    else:
        # this QUA variable name shouldn't interfere with commonly used QUA variable names, e.g. i = declare(int)
        _long_wait_loop_index = declare(int)

        loop_runs = (wait_time // threshold_for_looping) - 1

        # wait for N x threshold wait time
        with for_(_long_wait_loop_index, 0, _long_wait_loop_index < loop_runs, _long_wait_loop_index + 1):
            wait(threshold_for_looping, *elements)

        # wait for the remainder
        wait(int(wait_time - threshold_for_looping * loop_runs), *elements)



"""Tools to help handling units and conversions.

Content:
    - unit: class containing the relevant units.
    - to_clock_cycles: converts durations from ns to clock cycles.
    - demod2volts: converts demodulated data to volts.
    - raw2volts: converts raw data to volts.
"""

from qm import Program
from inspect import stack
from warnings import warn
from typing import Union
from numpy import round, ndarray, log10, sqrt


class _nanosecond:
    def __init__(self):
        pass

    def _get_value(self) -> float:
        for frameinfo in stack():
            for var in frameinfo.frame.f_locals.values():
                if isinstance(var, Program):  # we have a qua program being declared somewhere
                    if var._is_in_scope:  # this is set by __enter__ and unset by __exit__ of Program
                        return 0.25  # now it is in clock cycles
        else:
            return 1.0  # this is in nanoseconds

    def __mul__(self, other: Union[int, float]) -> float:
        value = self._get_value()

        return value * other

    def __rmul__(self, other: Union[int, float]) -> float:
        return self.__mul__(other)


class _hz:
    def __init__(self):
        pass

    def __mul__(self, other: Union[int, float]) -> float:
        return other

    def __rmul__(self, other: Union[int, float]) -> float:
        return self.__mul__(other)


class _ensure_integer(float):
    def __new__(cls, value: Union[int, float], verbose: bool = True, *args, **kwargs):
        cls.verbose = verbose
        return super(cls, cls).__new__(cls, value)

    def __mul__(self, other: Union[int, float]) -> int:
        result, remainder = divmod(float(self) * other, 1)

        if remainder != 0:
            if self.verbose:
                warn(
                    f"Warning: the specified duration ({other}) to be converted to clock cycles in not an integer. It has been converted to int ({result}) to avoid subsequent errors.",
                    RuntimeWarning,
                )

        return int(result)

    def __rmul__(self, other: Union[int, float]) -> int:
        return self.__mul__(other)


class _ensure_rounded_integer(float):
    def __new__(cls, value: Union[int, float], verbose: bool = True, *args, **kwargs):
        cls.verbose = verbose
        return super(cls, cls).__new__(cls, value)

    def __mul__(self, other: Union[int, float]) -> int:
        result, remainder = divmod(float(self) * other, 1)
        if remainder != 0:
            if self.verbose:
                warn(
                    f"Warning: the specified frequency ({other}) to be converted to Hz in not an integer. It has been converted to {int(round(result + remainder))} to avoid subsequent errors.",
                    RuntimeWarning,
                )

        return int(round(result + remainder))

    def __rmul__(self, other: Union[int, float]) -> int:
        return self.__mul__(other)


class unit:
    def __init__(self, coerce_to_integer=False, verbose=True):
        # Time units
        self._ns = _nanosecond()
        self.cc = 1 / 4

        # Handling time units
        self.ensure_integer = coerce_to_integer
        self.verbose = verbose

        # Frequency units
        self._Hz = _hz()
        self.mHz = 0.001

        # Voltage units
        self.V = 1
        self.mV = 1e-3
        self.uV = 1e-6

    @property
    def GHz(self) -> float:
        if self.ensure_integer:
            return _ensure_rounded_integer(1e9 * self._Hz, False)
        else:
            return 1e9 * self._Hz

    @property
    def MHz(self) -> float:
        if self.ensure_integer:
            return _ensure_rounded_integer(1e6 * self._Hz, False)
        else:
            return 1e6 * self._Hz

    @property
    def kHz(self) -> float:
        if self.ensure_integer:
            return _ensure_rounded_integer(1e3 * self._Hz, False)
        else:
            return 1e3 * self._Hz

    @property
    def Hz(self) -> float:
        if self.ensure_integer:
            return _ensure_rounded_integer(1 * self._Hz, False)
        else:
            return 1 * self._Hz

    @property
    def ns(self) -> float:
        if self.ensure_integer:
            return _ensure_integer(1 * self._ns, self.verbose)
        else:
            return 1 * self._ns

    @property
    def us(self) -> float:
        if self.ensure_integer:
            return _ensure_integer(1e3 * self._ns, self.verbose)
        else:
            return 1e3 * self._ns

    @property
    def ms(self) -> float:
        if self.ensure_integer:
            return _ensure_integer(1e6 * self._ns, self.verbose)
        else:
            return 1e6 * self._ns

    @property
    def s(self) -> float:
        if self.ensure_integer:
            return _ensure_integer(1e9 * self._ns, self.verbose)
        else:
            return 1e9 * self._ns

    @property
    def clock_cycle(self) -> float:
        if self.ensure_integer:
            return _ensure_integer(4 * self._ns, self.verbose)
        else:
            return 4 * self._ns

    # conversion functions
    def to_clock_cycles(self, t):
        """Converts a duration to clock cycles.

        :param t: duration in ns.
        :return: duration in clock cycles. It returns only the integer part (floor division).
        """
        if not float(t).is_integer():
            print(
                f"Warning: the specified duration ({t}) to be converted to clock cycles in not an integer. It has been converted to int ({int(t)}) to avoid subsequent errors."
            )
            t = int(t * self.ns)
        return int(t // 4)

    def demod2volts(
        self,
        data: Union[float, ndarray],
        duration: Union[float, int],
        single_demod: bool = False,
    ) -> Union[float, ndarray]:
        """Converts the demodulated data to volts.

        :param data: demodulated data. Must be a python variable or array.
        :param duration: demodulation duration in ns. **WARNING**: this must be the duration of one slice in the case of ```demod.sliced``` and ```demod.accumulated```.
        :param single_demod: Flag to add the additional factor of 2 needed for single demod.
        :return: the demodulated data in volts.
        """
        if single_demod:
            return 2 * 4096 * data * self.V / duration
        else:
            return 4096 * data * self.V / duration

    def volts2demod(
        self,
        value_in_volts: Union[float, ndarray],
        duration: Union[float, int],
        single_demod: bool = False,
    ) -> Union[float, ndarray]:
        """Converts the volts to demodulated data units.

        :param value_in_volts: some value in volts. Must be a python variable or array.
        :param duration: demodulation duration in ns.
        :param single_demod: Flag to add the additional factor of 2 needed for single demod.
        :return: the same value in demodulated data units.
        """
        if single_demod:
            return (value_in_volts * duration) / (2 * 4096 * self.V)
        else:
            return (value_in_volts * duration) / (4096 * self.V)

    def raw2volts(self, data: Union[float, ndarray]) -> Union[float, ndarray]:
        """Converts the raw data to volts.

        :param data: raw data. Must be a python variable or array.
        :return: the raw data in volts.
        """
        return data * self.V / 4096

    @staticmethod
    def volts2dBm(Vp: Union[float, ndarray], Z: Union[float, int] = 50) -> Union[float, ndarray]:
        """Converts the peak voltage (amplitude) from volts to dBm.

        :param Vp: the peak voltage (amplitude) in volts. Must be a python variable or array.
        :param Z: the impedance. Default is 50 ohms.
        :return: the power in dBm.
        """
        return 10 * log10(((Vp / sqrt(2)) ** 2 * 1000) / Z)

    @staticmethod
    def dBm2volts(P_dBm: Union[float, ndarray], Z: Union[float, int] = 50) -> Union[float, ndarray]:
        """Converts the power from dBm to volts (peak voltage or amplitude).

        :param P_dBm: the power in dBm. Must be a python variable or array.
        :param Z: the impedance. Default is 50 ohms.
        :return: the corresponding peak voltage (amplitude) in volts.
        """
        return sqrt(2 * Z / 1000) * 10 ** (P_dBm / 20)



import copy
import re
from collections import defaultdict
from typing import List

from matplotlib import patches
from matplotlib.axes import Axes
import matplotlib.colors as mcolors

from .layout import PORT_POSITIONS, PORT_SIZE


class PortAnnotation:
    def __init__(self, labels, color, con, slot, port, io_type, signal_type, instrument_id):
        self.labels = labels
        self.color = color
        self.con = con
        self.slot = slot
        self.port = port
        self.io_type = io_type
        self.signal_type = signal_type
        self.instrument_id = instrument_id

    def draw(self, ax: Axes):
        x, y = PORT_POSITIONS[self.instrument_id][self.signal_type][self.io_type][self.port - 1]
        fill_color = self.color if self.labels else "none"
        if self.signal_type == "digital":
            outline_colour = "whitesmoke"
        else:
            outline_colour = "dimgrey"

        if self.signal_type == "digital" and self.instrument_id in ["lf-fem", "mw-fem"]:
            port_size = PORT_SIZE / 2.4
            bbox = dict(facecolor=fill_color, alpha=0.8, edgecolor="none")
        elif self.instrument_id in ["opx+", "octave"]:
            port_size = PORT_SIZE
            port_label_distance = PORT_SIZE * 1.3
            ax.text(
                x - port_label_distance,
                y,
                str(self.port),
                ha="center",
                va="center",
                fontsize=8,
                fontweight="bold",
                color=outline_colour,
            )
            bbox = None
        elif self.instrument_id in ["external-mixer"]:
            port_size = PORT_SIZE * 2.4
            port_label_distance = PORT_SIZE * 3.2
            ax.text(
                x - port_label_distance,
                y,
                str(self.port),
                ha="center",
                va="center",
                fontsize=8,
                fontweight="bold",
                color=outline_colour,
            )
            bbox = None
        else:
            raise NotImplementedError(f"No port-annotation drawing for {self.instrument_id}")

        ax.add_patch(patches.Circle((x, y), port_size, edgecolor=outline_colour, facecolor=fill_color))
        labels = combine_labels_for_same_line_type(self.labels)

        # qubit line annotation
        ax.text(x, y, "\n".join(labels), ha="center", va="center", fontsize=9, color="black", bbox=bbox)
        ax.set_facecolor("lightgrey")

        if self.instrument_id == "octave" and self.signal_type == "analog":
            for port in [2 * self.port, 2 * self.port - 1]:
                if self.io_type == "output":
                    # draw inter-instrument port annotations for convenience
                    annotation_inter_output_I = copy.deepcopy(self)
                    annotation_inter_output_I.io_type = "inter_output"
                    annotation_inter_output_I.port = port
                    annotation_inter_output_I.draw(ax)
                elif self.io_type == "input" and self.port == 1:
                    annotation_inter_input_I = copy.deepcopy(self)
                    annotation_inter_input_I.io_type = "inter_input"
                    annotation_inter_input_I.port = port
                    annotation_inter_input_I.draw(ax)

    def title_axes(self, ax: Axes):
        if self.slot is not None:
            ax.set_title(f"{self.slot}: {self.instrument_id}", y=-0.1, va="bottom")


def get_contrast_color(color):
    """
    Returns black or white depending on the luminance of the input color.

    Parameters:
    color (str or tuple): The color to evaluate. Can be a string ('red', '#00FF00', etc.)
                          or an RGB tuple ((1, 0, 0), (0.5, 0.5, 0.5), etc.)

    Returns:
    str: 'black' if the input color is light, 'white' if it is dark.
    """
    # Convert color to RGB format if it's a string
    rgb = mcolors.to_rgb(color)

    # Calculate luminance
    luminance = 0.2126 * rgb[0] + 0.7152 * rgb[1] + 0.0722 * rgb[2]

    # Return 'black' if luminance is high (light color), 'white' if luminance is low (dark color)
    return "black" if luminance > 0.5 else "white"


def combine_labels_for_same_line_type(labels: List[str]):
    # Dictionary to group strings by line type
    grouped_lines = defaultdict(list)

    # Regular expression to parse the qubit line label
    pattern = re.compile(r"q(\d+)\.(.+)")

    # Parse each string and group by line type
    for s in labels:
        match = pattern.match(s)
        if match:
            index = int(match.group(1))  # Extract the qubit index
            line_type = match.group(2)  # Extract the line type
            grouped_lines[line_type].append(index)
        else:
            return labels

    # Result list to store the reduced strings
    reduced_strings = []
    # Process each line type group
    for line_type, indices in grouped_lines.items():
        indices.sort()
        ranges = []
        start = indices[0]
        end = start

        # Identify contiguous ranges of indices
        for i in range(1, len(indices)):
            if indices[i] == end + 1:
                # Extend the current range
                end = indices[i]
            else:
                # Save the current range and start a new one
                if start == end:
                    ranges.append(f"{start}")
                else:
                    ranges.append(f"{start}-{end}")
                start = indices[i]
                end = start

        # Append the final range or index
        if start == end:
            ranges.append(f"{start}")
        else:
            ranges.append(f"{start}-{end}")

        # Combine ranges into the final format
        reduced_strings.extend([f"q{range}.{line_type}" for range in ranges])

    return reduced_strings



import numpy as np

from qm.qua._dsl import QuaVariable, QuaExpression
from qm.qua import declare, assign, play, fixed, Cast, amp, wait, ramp, ramp_to_zero
from typing import Union, List, Dict


class VoltageGateSequence:
    def __init__(self, configuration: Dict, elements: List[str]):
        """
        Initializes a VirtualGateSequence object for designing arbitrary pulse sequences using virtual gates.

        This class allows the creation of complex pulse sequences by defining voltage levels (points) in gate-space
        and durations for multiple elements. It supports the addition of steps and ramps, and keeps track of average
        voltage levels for compensation pulses.

        The `configuration` provided will be updated to include necessary operations and waveforms for the sequence.

        :param configuration: A dictionary representing the OPX configuration (this will be modified)
        :param elements: A list of elements (strings) involved in the virtual gate operations.
        """
        # List of the elements involved in the virtual gates
        self._elements = elements
        # The OPX configuration
        self._config = configuration
        # Initialize the current voltage level for sticky elements
        self.current_level = [0.0 for _ in self._elements]
        # Relevant voltage points in the charge stability diagram
        self._voltage_points = {}
        # Keep track of the averaged voltage played for defining the compensation pulse at the end of the sequence
        self.average_power = [0 for _ in self._elements]
        self._expression = None
        self._expression2 = None
        # Add to the config the step operation (length=16ns & amp=0.25V)
        for el in self._elements:
            self._config["elements"][el]["operations"]["step"] = "step_pulse"
        self._config["pulses"]["step_pulse"] = {
            "operation": "control",
            "length": 16,
            "waveforms": {"single": "step_wf"},
        }
        self._config["waveforms"]["step_wf"] = {"type": "constant", "sample": 0.25}

    def _check_name(self, name, key):
        if name in key:
            return self._check_name(name + "%", key)
        else:
            return name

    def _add_op_to_config(self, el: str, name: str, amplitude: float, length: int) -> str:
        """Add an operation to an element when the amplitude is fixed to release the number of real-time operations on
        the OPX.

        :param el: the element to which we want to add the operation.
        :param name: name of the operation.
        :param amplitude: Amplitude of the pulse in V.
        :param length: Duration of the pulse in ns.
        :return : The name of the created operation.
        """
        op_name = self._check_name(name, self._config["elements"][el]["operations"])
        pulse_name = self._check_name(f"{el}_{op_name}_pulse", self._config["pulses"])
        wf_name = self._check_name(f"{el}_{op_name}_wf", self._config["waveforms"])
        self._config["elements"][el]["operations"][op_name] = pulse_name
        self._config["pulses"][pulse_name] = {
            "operation": "control",
            "length": length,
            "waveforms": {"single": wf_name},
        }
        self._config["waveforms"][wf_name] = {"type": "constant", "sample": amplitude}
        return op_name

    @staticmethod
    def _check_duration(duration: int):
        if duration is not None and not isinstance(duration, (QuaVariable, QuaExpression)):
            assert duration >= 4, "The duration must be a larger than 16 ns."

    def _update_averaged_power(self, level, duration, ramp_duration=None, current_level=None):
        if self.is_QUA(level):
            self._expression = declare(fixed)
            assign(self._expression, level)
            # Multiplication by 1024 to increase the resolution since duration*level must be an integer
            new_average = Cast.mul_int_by_fixed(duration << 10, self._expression)
        elif self.is_QUA(duration):
            new_average = Cast.mul_int_by_fixed(duration << 10, float(level))
        else:
            new_average = int(np.round((level * duration) * 1024))

        if ramp_duration is not None:
            if not self.is_QUA(ramp_duration):
                if self.is_QUA(level):
                    self._expression2 = declare(fixed)
                    assign(self._expression2, (self._expression + current_level) >> 1)
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, self._expression2)
                elif self.is_QUA(current_level):
                    expression2 = declare(fixed)
                    assign(expression2, (level + current_level) >> 1)
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, expression2)
                elif self.is_QUA(duration):
                    new_average += Cast.mul_int_by_fixed(ramp_duration << 10, (level + current_level) / 2)
                else:
                    new_average += int(np.round(1024 * (level + current_level) * ramp_duration / 2))

            else:
                pass
        return new_average

    @staticmethod
    def is_QUA(var):
        return isinstance(var, (QuaVariable, QuaExpression))

    def add_step(
        self,
        level: list[Union[int, QuaExpression, QuaVariable]] = None,
        duration: Union[int, QuaExpression, QuaVariable] = None,
        voltage_point_name: str = None,
        ramp_duration: Union[int, QuaExpression, QuaVariable] = None,
    ) -> None:
        """Add a voltage level to the pulse sequence.
        The voltage level is either identified by its voltage_point_name if added to the voltage_point dict beforehand, or by its level and duration.
        A ramp_duration can be used to ramp to the desired level instead of stepping to it.

        :param level: Desired voltage level of the different gates composing the virtual gate in Volt.
        :param duration: How long the voltage level should be maintained in ns. Must be a multiple of 4ns and larger than 16ns.
        :param voltage_point_name: Name of the voltage level if added to the list of relevant points in the charge stability map.
        :param ramp_duration: Duration in ns of the ramp if the voltage should be ramped to the desired level instead of stepped. Must be a multiple of 4ns and larger than 16ns.
        """
        self._check_duration(duration)
        self._check_duration(ramp_duration)
        if level is not None:
            if type(level) is not list or len(level) != len(self._elements):
                raise TypeError(
                    "the provided level must be a list of same length as the number of elements involved in the virtual gate."
                )

        if voltage_point_name is not None and duration is None:
            _duration = self._voltage_points[voltage_point_name]["duration"]
        elif duration is not None:
            _duration = duration
        else:
            raise RuntimeError(
                "Either the voltage_point_name or the duration and desired voltage level must be provided."
            )

        for i, gate in enumerate(self._elements):
            if voltage_point_name is not None and level is None:
                voltage_level = self._voltage_points[voltage_point_name]["coordinates"][i]
            elif level is not None:
                voltage_point_name = "unregistered_value"
                voltage_level = level[i]
            else:
                raise RuntimeError(
                    "Either the voltage_point_name or the duration and desired voltage level must be provided."
                )
            # Play a step
            if ramp_duration is None:
                self.average_power[i] += self._update_averaged_power(voltage_level, _duration)

                # Dynamic amplitude change...
                if self.is_QUA(voltage_level) or self.is_QUA(self.current_level[i]):
                    # if dynamic duration --> play step and wait
                    if self.is_QUA(_duration):
                        play("step" * amp((voltage_level - self.current_level[i]) * 4), gate)
                        wait((_duration - 16) >> 2, gate)
                    # if constant duration --> new operation and play(*amp(..))
                    else:
                        operation = self._add_op_to_config(
                            gate,
                            "step",
                            amplitude=0.25,
                            length=_duration,
                        )
                        play(operation * amp((voltage_level - self.current_level[i]) * 4), gate)

                # Fixed amplitude but dynamic duration --> new operation and play(duration=..)
                elif isinstance(_duration, (QuaVariable, QuaExpression)):
                    operation = self._add_op_to_config(
                        gate,
                        voltage_point_name,
                        amplitude=voltage_level - self.current_level[i],
                        length=16,
                    )
                    play(operation, gate, duration=_duration >> 2)

                # Fixed amplitude and duration --> new operation and play()
                else:
                    operation = self._add_op_to_config(
                        gate,
                        voltage_point_name,
                        amplitude=voltage_level - self.current_level[i],
                        length=_duration,
                    )
                    play(operation, gate)

            # Play a ramp
            else:
                self.average_power[i] += self._update_averaged_power(
                    voltage_level, _duration, ramp_duration, self.current_level[i]
                )

                if not self.is_QUA(ramp_duration):
                    ramp_rate = 1 / ramp_duration
                    play(ramp((voltage_level - self.current_level[i]) * ramp_rate), gate, duration=ramp_duration >> 2)
                    wait(_duration >> 2, gate)

            self.current_level[i] = voltage_level

    def add_compensation_pulse(self, duration: int) -> None:
        """Add a compensation pulse of the specified duration whose amplitude is derived from the previous operations.

        :param duration: Duration of the compensation pulse in clock cycles (4ns). Must be larger than 4 clock cycles.
        """
        self._check_duration(duration)
        for i, gate in enumerate(self._elements):
            if not self.is_QUA(self.average_power[i]):
                compensation_amp = -0.001 * self.average_power[i] / duration
                operation = self._add_op_to_config(
                    gate, "compensation", amplitude=compensation_amp - self.current_level[i], length=duration
                )
                play(operation, gate)
            else:
                operation = self._add_op_to_config(gate, "compensation", amplitude=0.25, length=duration)
                compensation_amp = declare(fixed)
                eval_average_power = declare(int)
                assign(eval_average_power, self.average_power[i])
                assign(compensation_amp, -Cast.mul_fixed_by_int(0.0009765625 / duration, eval_average_power))
                play(operation * amp((compensation_amp - self.current_level[i]) * 4), gate)
            self.current_level[i] = compensation_amp

    def ramp_to_zero(self, duration: int = None):
        """Ramp all the gate voltages down to zero Volt and reset the averaged voltage derived for defining the compensation pulse.

        :param duration: How long will it take for the voltage to ramp down to 0V in clock cycles (4ns). If not
            provided, the default pulse duration defined in the configuration will be used.
        """
        for i, gate in enumerate(self._elements):
            ramp_to_zero(gate, duration)
            self.current_level[i] = 0
            self.average_power[i] = 0
        if self._expression is not None:
            assign(self._expression, 0)
        if self._expression2 is not None:
            assign(self._expression2, 0)

    def add_points(self, name: str, coordinates: list, duration: int) -> None:
        """Register a relevant voltage point.

        :param name: Name of the voltage point.
        :param coordinates: Voltage value of each gate involved in the virtual gate in V.
        :param duration: How long should the voltages be maintained at this level in ns. Must be larger than 16ns and a multiple of 4ns.
        """
        self._voltage_points[name] = {}
        self._voltage_points[name]["coordinates"] = coordinates
        self._voltage_points[name]["duration"] = duration



algebra_supported_types = (int, float)


class Parameter(object):
    def __init__(self, name: str):
        """An object to represents a simple parameter or any setter.
        :param name: Name of the parameter
        :type name: str
        """
        self.is_set = False
        self._value = None
        self.name = name

    def __call__(self, *args, **kwargs):
        if self.is_set:
            if callable(self._value):
                return self._value(*args, **kwargs)
            else:
                return self._value
        else:
            raise AssertionError("Parameter {} is not set".format(self.name))

    @property
    def value(self):
        """Returns the value of the parameter"""
        return self._value

    @value.setter
    def value(self, v):
        self.is_set = True
        self._value = v

    @property
    def len(self):
        """Returns a new parameter with a value equal to the length of the current parameter."""
        len_self = Parameter("len_" + self.name)

        def func():
            if self.is_set:
                if callable(self._value):
                    return len(self._value())
                else:
                    return len(self._value)
            else:
                raise AssertionError("Parameter {} is not set".format(self.name))

        len_self._value = func
        len_self.is_set = True
        return len_self

    def __add__(self, other):
        if isinstance(other, Parameter):
            new_param = Parameter("add_" + self.name + "_" + other.name)

            def func():
                if self.is_set and other.is_set:
                    return self() + other()
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))
                elif not other.is_set:
                    raise AssertionError("Parameter {} is not set".format(other.name))

            new_param._value = func
            new_param.is_set = True
            return new_param
        elif isinstance(other, algebra_supported_types):
            new_param = Parameter("add_" + self.name + "_" + str(other))

            def func():
                if self.is_set:
                    return self() + other
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))

            new_param._value = func
            new_param.is_set = True
            return new_param

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Parameter):
            new_param = Parameter("sub_ " + self.name + "_" + other.name)

            def func():
                if self.is_set and other.is_set:
                    return self() - other()
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))
                elif not other.is_set:
                    raise AssertionError("Parameter {} is not set".format(other.name))

            new_param._value = func
            new_param.is_set = True
            return new_param
        elif isinstance(other, algebra_supported_types):
            new_param = Parameter("sub_" + self.name + "_" + str(other))

            def func():
                if self.is_set:
                    return self() - other
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))

            new_param._value = func
            new_param.is_set = True
            return new_param

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, Parameter):
            new_param = Parameter("mul_ " + self.name + "_" + other.name)

            def func():
                if self.is_set and other.is_set:
                    return self() * other()
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))
                elif not other.is_set:
                    raise AssertionError("Parameter {} is not set".format(other.name))

            new_param._value = func
            new_param.is_set = True
            return new_param
        elif isinstance(other, algebra_supported_types):
            new_param = Parameter("mul_" + self.name + "_" + str(other))

            def func():
                if self.is_set:
                    return self() * other
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))

            new_param._value = func
            new_param.is_set = True
            return new_param

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Parameter):
            new_param = Parameter("div_ " + self.name + "_" + other.name)

            def func():
                if self.is_set and other.is_set:
                    return self() / other()
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))
                elif not other.is_set:
                    raise AssertionError("Parameter {} is not set".format(other.name))

            new_param._value = func
            new_param.is_set = True
            return new_param
        elif isinstance(other, algebra_supported_types):
            new_param = Parameter("div_" + self.name + "_" + str(other))

            def func():
                if self.is_set:
                    return self() / other
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))

            new_param._value = func
            new_param.is_set = True
            return new_param

    def __rtruediv__(self, other):
        return self.__truediv__(other)

    def __pow__(self, other):
        if isinstance(other, Parameter):
            new_param = Parameter("pow_ " + self.name + "_" + other.name)

            def func():
                if self.is_set and other.is_set:
                    return self() ** other()
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))
                elif not other.is_set:
                    raise AssertionError("Parameter {} is not set".format(other.name))

            new_param._value = func
            new_param.is_set = True
            return new_param
        elif isinstance(other, algebra_supported_types):
            new_param = Parameter("pow_" + self.name + "_" + str(other))

            def func():
                if self.is_set:
                    return self() ** other
                elif not self.is_set:
                    raise AssertionError("Parameter {} is not set".format(self.name))

            new_param._value = func
            new_param.is_set = True
            return new_param

    def __rpow__(self, other):
        return self.__pow__(other)


class ConfigVars(object):
    def __init__(self):
        """An object that holds a collection of Parameters. Useful to write parametric
        QUA configurations.
        """
        self.params = {}

    def parameter(self, name: str, setter=None):
        """Returns a Parameter with the given name
        :param name: Name of the parameter
        :type name: str
        :param setter: any function
        :type setter: Callable
        """
        if name.find(" ") != -1:
            raise ValueError("Parameter name cannot contain spaces. " "Use underscore of camelCase.")
        if name not in self.params.keys():
            self.params[name] = Parameter(name)
        if setter is not None:
            self.params[name].value = setter
        return self.params[name]

    def set(self, **kwargs):
        """sets the parameters with the given values specified as a dictionary"""
        for key, value in kwargs.items():
            if key not in self.params.keys():
                self.params[key] = Parameter(key)
            self.params[key].value = value


def _is_parametric(obj):
    if callable(obj):
        try:
            obj()
            return False
        except AssertionError:
            return True
        except TypeError:
            return False
    elif hasattr(obj, "__dict__"):
        for elm in obj.__dict__:
            _obj = getattr(obj, elm)
            if hasattr(_obj, "__dict__"):
                for k, v in _obj.__dict__.items():
                    if _is_parametric(v):
                        return True
            elif isinstance(_obj, dict):
                for k, v in _obj.items():
                    if _is_parametric(v):
                        return True
            elif isinstance(_obj, list):
                for elm in _obj:
                    if _is_parametric(elm):
                        return True
            elif isinstance(_obj, tuple):
                for elm in list(_obj):
                    if _is_parametric(elm):
                        return True
            else:
                if _is_parametric(_obj):
                    return True
        return False
    elif isinstance(obj, dict):
        for k, v in obj.items():
            if _is_parametric(v):
                return True
    elif isinstance(obj, str):
        return False
    elif obj is None:
        return False
    elif isinstance(obj, list):
        for elm in obj:
            if _is_parametric(elm):
                return True
    elif isinstance(obj, tuple):
        for elm in list(obj):
            if _is_parametric(elm):
                return True
    else:
        return False
    return False


def _is_callable(obj):
    if callable(obj):
        return True
    elif isinstance(obj, dict):
        for k, v in obj.items():
            if _is_callable(v):
                return True
        return False
    elif isinstance(obj, list):
        for e in obj:
            if _is_callable(e):
                return True
        return False
    elif isinstance(obj, tuple):
        return _is_callable(list(obj))
    else:
        return False
    return False



from qualang_tools.loops.loops import from_array
from qualang_tools.loops.loops import (
    qua_arange,
    qua_linspace,
    qua_logspace,
    get_equivalent_log_array,
)

__all__ = [
    "from_array",
    "qua_arange",
    "qua_linspace",
    "qua_logspace",
    "get_equivalent_log_array",
]



import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import minimize


def _false_detections(threshold, Ig, Ie):
    if np.mean(Ig) < np.mean(Ie):
        false_detections_var = np.sum(Ig > threshold) + np.sum(Ie < threshold)
    else:
        false_detections_var = np.sum(Ig < threshold) + np.sum(Ie > threshold)
    return false_detections_var


def two_state_discriminator(Ig, Qg, Ie, Qe, b_print=True, b_plot=True):
    """
    Given two blobs in the IQ plane representing two states, finds the optimal threshold to discriminate between them
    and calculates the fidelity. Also returns the angle in which the data needs to be rotated in order to have all the
    information in the `I` (`X`) axis.

    .. note::
        This function assumes that there are only two blobs in the IQ plane representing two states (ground and excited)
        Unexpected output will be returned in other cases.


    :param float Ig: A vector containing the `I` quadrature of data points in the ground state
    :param float Qg: A vector containing the `Q` quadrature of data points in the ground state
    :param float Ie: A vector containing the `I` quadrature of data points in the excited state
    :param float Qe: A vector containing the `Q` quadrature of data points in the excited state
    :param bool b_print: When true (default), prints the results to the console.
    :param bool b_plot: When true (default), plots the results in a new figure.
    :returns: A tuple of (angle, threshold, fidelity, gg, ge, eg, ee).
        angle - The angle (in radians) in which the IQ plane has to be rotated in order to have all the information in
            the `I` axis.
        threshold - The threshold in the rotated `I` axis. The excited state will be when the `I` is larger (>) than
            the threshold.
        fidelity - The fidelity for discriminating the states.
        gg - The matrix element indicating a state prepared in the ground state and measured in the ground state.
        ge - The matrix element indicating a state prepared in the ground state and measured in the excited state.
        eg - The matrix element indicating a state prepared in the excited state and measured in the ground state.
        ee - The matrix element indicating a state prepared in the excited state and measured in the excited state.
    """
    # Condition to have the Q equal for both states:
    angle = np.arctan2(np.mean(Qe) - np.mean(Qg), np.mean(Ig) - np.mean(Ie))
    C = np.cos(angle)
    S = np.sin(angle)
    # Condition for having e > Ig
    if np.mean((Ig - Ie) * C - (Qg - Qe) * S) > 0:
        angle += np.pi
        C = np.cos(angle)
        S = np.sin(angle)

    Ig_rotated = Ig * C - Qg * S
    Qg_rotated = Ig * S + Qg * C

    Ie_rotated = Ie * C - Qe * S
    Qe_rotated = Ie * S + Qe * C

    fit = minimize(
        _false_detections,
        0.5 * (np.mean(Ig_rotated) + np.mean(Ie_rotated)),
        (Ig_rotated, Ie_rotated),
        method="Nelder-Mead",
    )
    threshold = fit.x[0]

    gg = np.sum(Ig_rotated < threshold) / len(Ig_rotated)
    ge = np.sum(Ig_rotated > threshold) / len(Ig_rotated)
    eg = np.sum(Ie_rotated < threshold) / len(Ie_rotated)
    ee = np.sum(Ie_rotated > threshold) / len(Ie_rotated)

    fidelity = 100 * (gg + ee) / 2

    if b_print:
        print(
            f"""
        Fidelity Matrix:
        -----------------
        | {gg:.3f} | {ge:.3f} |
        ----------------
        | {eg:.3f} | {ee:.3f} |
        -----------------
        IQ plane rotated by: {180 / np.pi * angle:.1f}{chr(176)}
        Threshold: {threshold:.3e}
        Fidelity: {fidelity:.1f}%
        """
        )

    if b_plot:
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2)
        ax1.plot(Ig, Qg, ".", alpha=0.1, label="Ground", markersize=2)
        ax1.plot(Ie, Qe, ".", alpha=0.1, label="Excited", markersize=2)
        ax1.axis("equal")
        ax1.legend(["Ground", "Excited"])
        ax1.set_xlabel("I")
        ax1.set_ylabel("Q")
        ax1.set_title("Original Data")

        ax2.plot(Ig_rotated, Qg_rotated, ".", alpha=0.1, label="Ground", markersize=2)
        ax2.plot(Ie_rotated, Qe_rotated, ".", alpha=0.1, label="Excited", markersize=2)
        ax2.axis("equal")
        ax2.set_xlabel("I")
        ax2.set_ylabel("Q")
        ax2.set_title("Rotated Data")

        ax3.hist(Ig_rotated, bins=50, alpha=0.75, label="Ground")
        ax3.hist(Ie_rotated, bins=50, alpha=0.75, label="Excited")
        ax3.axvline(x=threshold, color="k", ls="--", alpha=0.5)
        text_props = dict(
            horizontalalignment="center",
            verticalalignment="center",
            transform=ax3.transAxes,
        )
        ax3.text(0.7, 0.9, f"{threshold:.3e}", text_props)
        ax3.set_xlabel("I")
        ax3.set_title("1D Histogram")

        ax4.imshow(np.array([[gg, ge], [eg, ee]]))
        ax4.set_xticks([0, 1])
        ax4.set_yticks([0, 1])
        ax4.set_xticklabels(labels=["|g>", "|e>"])
        ax4.set_yticklabels(labels=["|g>", "|e>"])
        ax4.set_ylabel("Prepared")
        ax4.set_xlabel("Measured")
        ax4.text(0, 0, f"{100 * gg:.1f}%", ha="center", va="center", color="k")
        ax4.text(1, 0, f"{100 * ge:.1f}%", ha="center", va="center", color="w")
        ax4.text(0, 1, f"{100 * eg:.1f}%", ha="center", va="center", color="w")
        ax4.text(1, 1, f"{100 * ee:.1f}%", ha="center", va="center", color="k")
        ax4.set_title("Fidelities")
        fig.tight_layout()

    return angle, threshold, fidelity, gg, ge, eg, ee



from .channel_spec import ChannelSpec
from .types import QubitsType, QubitPairsType
from .wiring_spec import WiringFrequency, WiringIOType, WiringLineType
from .connectivity_base import ConnectivityBase


class Connectivity(ConnectivityBase):
    def add_fixed_transmons(self, qubits: QubitsType):
        self.add_resonator_line(qubits)
        self.add_qubit_drive_lines(qubits)

    def add_flux_tunable_transmons(self, qubits: QubitsType):
        self.add_resonator_line(qubits)
        self.add_qubit_drive_lines(qubits)
        self.add_qubit_flux_lines(qubits)

    def add_resonator_line(self, qubits: QubitsType, triggered: bool = False, constraints: ChannelSpec = None):
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.RF,
            WiringIOType.INPUT_AND_OUTPUT,
            WiringLineType.RESONATOR,
            triggered,
            constraints,
            elements,
            shared_line=True,
        )

    def add_qubit_drive_lines(self, qubits: QubitsType, triggered: bool = False, constraints: ChannelSpec = None):
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.RF, WiringIOType.OUTPUT, WiringLineType.DRIVE, triggered, constraints, elements
        )

    def add_qubit_charge_lines(self, qubits: QubitsType, constraints: ChannelSpec = None):
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, WiringLineType.CHARGE, False, constraints, elements
        )

    def add_qubit_flux_lines(self, qubits: QubitsType, constraints: ChannelSpec = None):
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, WiringLineType.FLUX, False, constraints, elements
        )

    def add_qubit_pair_flux_lines(self, qubit_pairs: QubitPairsType, constraints: ChannelSpec = None):
        elements = self._make_qubit_pair_elements(qubit_pairs)
        return self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, WiringLineType.COUPLER, False, constraints, elements
        )

    def add_qubit_pair_cross_resonance_lines(
        self, qubit_pairs: QubitPairsType, triggered: bool = False, constraints: ChannelSpec = None
    ):
        elements = self._make_qubit_pair_elements(qubit_pairs)
        return self.add_wiring_spec(
            WiringFrequency.RF, WiringIOType.OUTPUT, WiringLineType.CROSS_RESONANCE, triggered, constraints, elements
        )

    def add_qubit_pair_zz_drive_lines(
        self, qubit_pairs: QubitPairsType, triggered: bool = False, constraints: ChannelSpec = None
    ):
        elements = self._make_qubit_pair_elements(qubit_pairs)
        return self.add_wiring_spec(
            WiringFrequency.RF, WiringIOType.OUTPUT, WiringLineType.ZZ_DRIVE, triggered, constraints, elements
        )



from qualang_tools.config.integration_weights_tools import (
    convert_integration_weights,
    compress_integration_weights,
    plot_integration_weights,
)
from qualang_tools.config.waveform_tools import (
    drag_gaussian_pulse_waveforms,
    drag_cosine_pulse_waveforms,
)
from qualang_tools.config.builder import ConfigBuilder
from qualang_tools.config.components import *
from qualang_tools.config.primitive_components import *
from qualang_tools.config.parameters import Parameter, ConfigVars
from qualang_tools.config.helper_tools import QuaConfig

__all__ = [
    "drag_gaussian_pulse_waveforms",
    "drag_cosine_pulse_waveforms",
    "convert_integration_weights",
    "compress_integration_weights",
    "plot_integration_weights",
    "Controller",
    "ArbitraryWaveform",
    "ConstantWaveform",
    "DigitalWaveform",
    "MeasurePulse",
    "ControlPulse",
    "Mixer",
    "Element",
    "MeasureElement",
    "ConstantIntegrationWeights",
    "ArbitraryIntegrationWeights",
    "PiecewiseConstantIntegrationWeights",
    "ElementCollection",
    "ReadoutResonator",
    "Transmon",
    "FluxTunableTransmon",
    "Coupler",
    "Oscillator",
    "Port",
    "AnalogInputPort",
    "AnalogOutputPort",
    "DigitalInputPort",
    "DigitalOutputPort",
    "Waveform",
    "Pulse",
    "Operation",
    "IntegrationWeights",
    "Weights",
    "DigitalSample",
    "Matrix2x2",
    "MixerData",
    "AnalogOutputFilter",
    "ConfigBuilder",
    "Parameter",
    "ConfigVars",
    "QuaConfig",
]



from qualang_tools.plot.plot import *
from qualang_tools.plot.fitting import *


__all__ = [
    "interrupt_on_close",
    "Fit",
    "Read",
    "get_simulated_samples_by_element",
    "plot_simulator_output",
    "plot_demodulated_data_1d",
    "plot_demodulated_data_2d",
]



from .data_folder_tools import *
from . import data_processors
from .data_processors import DEFAULT_DATA_PROCESSORS
from .data_handler import *

__all__ = [*data_folder_tools.__all__, data_processors, DEFAULT_DATA_PROCESSORS, *data_handler.__all__]
